/****************************************************************************
**
** This file is part of a Qt Solutions component.
** 
** Copyright (c) 2009 Nokia Corporation and/or its subsidiary(-ies).
** 
** Contact:  Qt Software Information (qt-info@nokia.com)
** 
** Commercial Usage  
** Licensees holding valid Qt Commercial licenses may use this file in
** accordance with the Qt Solutions Commercial License Agreement provided
** with the Software or, alternatively, in accordance with the terms
** contained in a written agreement between you and Nokia.
** 
** GNU Lesser General Public License Usage
** Alternatively, this file may be used under the terms of the GNU Lesser
** General Public License version 2.1 as published by the Free Software
** Foundation and appearing in the file LICENSE.LGPL included in the
** packaging of this file.  Please review the following information to
** ensure the GNU Lesser General Public License version 2.1 requirements
** will be met: http://www.gnu.org/licenses/old-licenses/lgpl-2.1.html.
** 
** In addition, as a special exception, Nokia gives you certain
** additional rights. These rights are described in the Nokia Qt LGPL
** Exception version 1.0, included in the file LGPL_EXCEPTION.txt in this
** package.
** 
** GNU General Public License Usage 
** Alternatively, this file may be used under the terms of the GNU
** General Public License version 3.0 as published by the Free Software
** Foundation and appearing in the file LICENSE.GPL included in the
** packaging of this file.  Please review the following information to
** ensure the GNU General Public License version 3.0 requirements will be
** met: http://www.gnu.org/copyleft/gpl.html.
** 
** Please note Third Party Software included with Qt Solutions may impose
** additional restrictions and it is the user's responsibility to ensure
** that they have met the licensing requirements of the GPL, LGPL, or Qt
** Solutions Commercial license and the relevant license of the Third
** Party Software they are using.
** 
** If you are unsure which license is appropriate for your use, please
** contact the sales department at qt-sales@nokia.com.
** 
****************************************************************************/



#include "qabstractstategroup.h"
#include "qabstractstategroup_p.h"
#include "qhistorystate.h"
#include "qhistorystate_p.h"
#include "qabstracttransition.h"
#include "qabstracttransition_p.h"
#include "qstatemachine.h"
#include "qstatemachine_p.h"
#include "qsignaltransition.h"
#include "qstatefinishedtransition.h"
#include "qstateaction.h"
#include "qstateaction_p.h"

#ifndef QT_NO_ANIMATION
#include "qanimationstate.h"
#include "qabstractanimation.h"
#endif

QT_BEGIN_NAMESPACE

/*!
  \class QtAbstractStateGroup

  \brief The QtAbstractStateGroup class is the base class of states that can have children.

  

  The QtAbstractStateGroup class is the abstract base class of states that can
  have child states. QtAbstractStateGroup is part of \l{The State Machine
  Framework}.

  The addTransition() function adds a transition. The removeTransition()
  function removes a transition.

  \code
  QtAbstractStateGroup *s1 = ...;
  QtAbstractState *s2 = ...;
  QPushButton button;
  s1->addTransition(&button, SIGNAL(clicked()), s2);
  \endcode

  The addHistoryState() function adds a history state.

  You can add actions to a state with the addEntryAction() and addExitAction()
  functions. The state executes the actions when the state is entered and
  exited, respectively.

  QtState provides built-in actions for setting properties and invoking methods
  of QObjects. The setPropertyOnEntry() and setPropertyOnExit() functions are
  used for defining property assignments that should be performed when a state
  is entered and exited, respectively.

  \code
  QLabel label;
  QtStateMachine machine;
  QtState *s1 = new QtState();
  s1->setPropertyOnEntry(&label, "text"; "Entered state s1");
  machine.addState(s1);
  \endcode

  The invokeMethodOnEntry() and invokeMethodOnExit() functions are used for
  defining method invocations that should be performed when a state is entered
  and exited, respectively.

  \code
  QtState *s2 = new QtState();
  s2->invokeMethodOnEntry(&label, "showMaximized");
  machine.addState(s2);
  \endcode
*/

/*!
    \enum QtAbstractStateGroup::HistoryType

    This enum specifies the type of history that a QtHistoryState records.

    \value ShallowHistory Only the immediate child states of the parent state
    are recorded. In this case a transition with the history state as its
    target will end up in the immediate child state that the parent was in the
    last time it was exited. This is the default.

    \value DeepHistory Nested states are recorded. In this case a transition
    with the history state as its target will end up in the most deeply nested
    descendant state the parent was in the last time it was exited.
*/

/*!
   \enum QtAbstractStateGroup::RestorePolicy

   This enum specifies the restore policy type for a state. The restore policy takes effect when
   the machine enters a state which has entry actions of the type QtStateSetPropertyAction. If the 
   restore policy of the state is set to RestoreProperties, the state machine will save the 
   value of the property before the QtStateSetPropertyAction is executed. 

   Later, when the machine either enters a state which has its restore policy set to 
   DoNotRestoreProperties or when it enters a state which does not set a value for the given 
   property, the property will automatically be restored to its initial value. The state machine
   will only detect which properties are being set if they are being set using a
   QtStateSetPropertyAction object set as entry action on a state.

   Special rules apply when using QtAnimationState. If a QtAnimationState registers that a property
   should be restored before entering the target state of its QtStateFinishedTransition, it will
   restore this property using a QtPropertyAnimation. 
   
   Only one initial value will be saved for any given property. If a value for a property has 
   already been saved by the state machine, it will not be overwritten until the property has been
   successfully restored. Once the property has been restored, the state machine will clear the 
   initial value until it enters a new state which sets the property and which has RestoreProperties
   as its restore policy.

   \value GlobalRestorePolicy The restore policy for the state should be retrieved using 
          QtStateMachine::globalRestorePolicy()
   \value DoNotRestoreProperties The state machine should not save the initial values of properties 
          set in the state and restore them later.
   \value RestoreProperties The state machine should save the initial values of properties 
          set in the state and restore them later.


   \sa setRestorePolicy(), restorePolicy(), addEntryAction(), setPropertyOnEntry()

*/

QtAbstractStateGroupPrivate::QtAbstractStateGroupPrivate()
    : restorePolicy(QtAbstractStateGroup::GlobalRestorePolicy),
      hasUnregisteredTransitions(false), errorState(0)
{
}

QtAbstractStateGroupPrivate::~QtAbstractStateGroupPrivate()
{
    qDeleteAll(entryActions);
    qDeleteAll(exitActions);
}

QtAbstractStateGroupPrivate *QtAbstractStateGroupPrivate::get(QtAbstractStateGroup *q)
{
    return q->d_func();
}

const QtAbstractStateGroupPrivate *QtAbstractStateGroupPrivate::get(const QtAbstractStateGroup *q)
{
    return q->d_func();
}

QList<QtAbstractState*> QtAbstractStateGroupPrivate::childStates() const
{
    QList<QtAbstractState*> result;
    QList<QObject*>::const_iterator it;
#ifdef QT_STATEMACHINE_SOLUTION
    const QObjectList &children = q_func()->children();
#endif
    for (it = children.constBegin(); it != children.constEnd(); ++it) {
        QtAbstractState *s = qobject_cast<QtAbstractState*>(*it);
        if (!s || qobject_cast<QtHistoryState*>(s))
            continue;
        result.append(s);
    }
    return result;
}

QList<QtHistoryState*> QtAbstractStateGroupPrivate::historyStates() const
{
    QList<QtHistoryState*> result;
    QList<QObject*>::const_iterator it;
#ifdef QT_STATEMACHINE_SOLUTION
    const QObjectList &children = q_func()->children();
#endif
    for (it = children.constBegin(); it != children.constEnd(); ++it) {
        QtHistoryState *h = qobject_cast<QtHistoryState*>(*it);
        if (h)
            result.append(h);
    }
    return result;
}

QList<QtAbstractTransition*> QtAbstractStateGroupPrivate::transitions() const
{
    QList<QtAbstractTransition*> result;
    QList<QObject*>::const_iterator it;
#ifdef QT_STATEMACHINE_SOLUTION
    const QObjectList &children = q_func()->children();
#endif
    for (it = children.constBegin(); it != children.constEnd(); ++it) {
        QtAbstractTransition *t = qobject_cast<QtAbstractTransition*>(*it);
        if (t)
            result.append(t);
    }
    return result;
}

/*!
  Constructs a new state group with the given \a parent state.
*/
QtAbstractStateGroup::QtAbstractStateGroup(QtAbstractStateGroup *parent)
    : QtAbstractState(*new QtAbstractStateGroupPrivate, parent)
{
}

/*!
  \internal
*/
QtAbstractStateGroup::QtAbstractStateGroup(QtAbstractStateGroupPrivate &dd,
                                         QtAbstractStateGroup *parent)
    : QtAbstractState(dd, parent)
{
}

/*!
  Destroys this state group.
*/
QtAbstractStateGroup::~QtAbstractStateGroup()
{
}

/*!
  Returns this state group's error state. 

  \sa QtStateMachine::errorState(), QtStateMachine::setErrorState()
*/
QtAbstractState *QtAbstractStateGroup::errorState() const
{
    Q_D(const QtAbstractStateGroup);
    return d->errorState;
}

/*!
  Sets this state group's error state to be the given \a state. If the error state is not set,
  or if it is set to 0, the state will inherit its parent's error state recursively.

  \sa QtStateMachine::setErrorState(), QtStateMachine::errorState()
*/
void QtAbstractStateGroup::setErrorState(QtAbstractState *state)
{
    Q_D(QtAbstractStateGroup);
    if (state != 0 && QtAbstractStatePrivate::get(state)->machine() != d->machine()) {
        qWarning("QtAbstractStateGroup::setErrorState: error state cannot belong "
                 "to a different state machine");
        return;
    }

    d->errorState = state;
}

/*!
  Adds the given \a transition. The transition has this state as the source,
  and the given \a target as the target state.

  \sa removeTransition()
*/
void QtAbstractStateGroup::addTransition(QtAbstractTransition *transition,
                                        QtAbstractState *target)
{
    addTransition(transition, QList<QtAbstractState*>() << target);
}

/*!
  Adds the given \a transition. The transition has this state as the source,
  and the given \a targets as the target states.
*/
void QtAbstractStateGroup::addTransition(QtAbstractTransition *transition,
                                        const QList<QtAbstractState*> &targets)
{
    Q_D(QtAbstractStateGroup);
    for (int i = 0; i < targets.size(); ++i) {
        if (!targets.at(i)) {
            qWarning("QtAbstractState::addTransition: cannot add transition to null state");
            return;
        }
        if (QtAbstractStatePrivate::get(targets.at(i))->machine() != d->machine()) {
            qWarning("QtAbstractState::addTransition: cannot add transition "
                     "to a state in a different state machine");
            return;
        }
    }
    Q_ASSERT(transition->parent() == 0);
    Q_ASSERT(QtAbstractTransitionPrivate::get(transition)->targetStates.isEmpty());
    transition->setParent(this);
    QtAbstractTransitionPrivate::get(transition)->targetStates = targets;
    d->hasUnregisteredTransitions = true;
}

/*!
  Adds a transition associated with the given \a signal of the given \a sender
  object. The transition has this state as the source, and the given \a target
  as the target state.
*/
void QtAbstractStateGroup::addTransition(const QObject *sender, const char *signal,
                                        QtAbstractState *target)
{
    addTransition(new QtSignalTransition(sender, signal), target);
}

/*!
  Adds the given \a transition. The transition has this state as the source,
  and has no target.
*/
void QtAbstractStateGroup::addTransition(QtAbstractTransition *transition)
{
    addTransition(transition, QList<QtAbstractState*>());
}

/*!
  Adds a transition that's triggered by the finished event of this state, and
  that has the given \a target state.

  \sa QtStateFinishedEvent
*/
void QtAbstractStateGroup::addFinishedTransition(QtAbstractState *target)
{
    addTransition(new QtStateFinishedTransition(this), target);
}

namespace {

class UnconditionalTransition : public QtAbstractTransition
{
protected:
    void onTransition() {}
    bool testCondition(QEvent *) const { return true; }
};

} // namespace

/*!
  Adds an unconditional transition from this state to the given \a target
  state.
*/
void QtAbstractStateGroup::addTransition(QtAbstractState *target)
{
    addTransition(new UnconditionalTransition(), target);
}

/*!
  Removes the given \a transition from this state.  The state releases
  ownership of the transition.

  \sa addTransition()
*/
void QtAbstractStateGroup::removeTransition(QtAbstractTransition *transition)
{
    if (!transition) {
        qWarning("QtAbstractState::removeTransition: cannot remove null transition");
        return;
    }
    if (QtAbstractTransitionPrivate::get(transition)->sourceState() != this) {
        qWarning("QtAbstractState::removeTransition: transition %p's source state (%p)"
                 " is different from this state (%p)",
                 transition, QtAbstractTransitionPrivate::get(transition)->sourceState(), this);
        return;
    }
    transition->setParent(0);
    QtAbstractTransitionPrivate::get(transition)->targetStates.clear();
}

/*!
  Creates a history state of the given \a type for this state and returns the
  new state.  The history state becomes a child of this state.
*/
QtHistoryState *QtAbstractStateGroup::addHistoryState(HistoryType type)
{
    return QtHistoryStatePrivate::create(type, this);
}

/*!
  Instructs this QtState to set the property with the given \a name of the
  given \a object to the given \a value when the state is entered. This 
  function will create an object of QtStateSetPropertyAction and add it to 
  the entry actions of the state.

  \sa setPropertyOnExit(), invokeMethodOnEntry(), addEntryAction()
*/
void QtAbstractStateGroup::setPropertyOnEntry(QObject *object, const char *name,
                                             const QVariant &value)
{
    addEntryAction(new QtStateSetPropertyAction(object, name, value));
}

/*!
  Instructs this QtState to set the property with the given \a name of the
  given \a object to the given \a value when the state is exited. This 
  function will create an object of QtStateSetPropertyAction and add it to
  the exit actions of the state.

  \sa setPropertyOnEntry(), invokeMethodOnExit(), addExitAction()
*/
void QtAbstractStateGroup::setPropertyOnExit(QObject *object, const char *name,
                                            const QVariant &value)
{
    addExitAction(new QtStateSetPropertyAction(object, name, value));
}

/*!
  Instructs this QtState to invoke the given \a method of the given \a object
  with the given \a arguments when the state is entered. This function will 
  create an object of QtStateInvokeMethodAction and add it to the entry actions 
  of the state.

  \sa invokeMethodOnExit(), setPropertyOnEntry(), addEntryAction()
*/
void QtAbstractStateGroup::invokeMethodOnEntry(QObject *object, const char *method,
                                              const QList<QVariant> &arguments)
{
    addEntryAction(new QtStateInvokeMethodAction(object, method, arguments));
}

/*!
  Instructs this QtState to invoke the given \a method of the given \a object
  with the given \a arguments when the state is exited. This function will create
  an object of QtStateInvokeMethodAction and add it to the exit actions of the state.

  \sa invokeMethodOnEntry(), setPropertyOnExit(), addExitAction()
*/
void QtAbstractStateGroup::invokeMethodOnExit(QObject *object, const char *method,
                                             const QList<QVariant> &arguments)
{
    addExitAction(new QtStateInvokeMethodAction(object, method, arguments));
}

/*!
  Adds the given \a action to this state. The action will be executed when
  this state is entered. The state takes ownership of the action.

  \sa addExitAction()
*/
void QtAbstractStateGroup::addEntryAction(QtStateAction *action)
{
    Q_D(QtAbstractStateGroup);
    if (action)
        d->entryActions.append(action);
}

/*!
  Adds the given \a action to this state. The action will be executed when
  this state is exited. The state takes ownership of the action.

  \sa addEntryAction()
*/
void QtAbstractStateGroup::addExitAction(QtStateAction *action)
{
    Q_D(QtAbstractStateGroup);
    if (action)
        d->exitActions.append(action);
}

/*!
  \reimp
*/
void QtAbstractStateGroup::onEntry()
{
    Q_D(QtAbstractStateGroup);
    for (int i = 0; i < d->entryActions.size(); ++i)
        QtStateActionPrivate::get(d->entryActions.at(i))->callExecute();
}

/*!
  \reimp
*/
void QtAbstractStateGroup::onExit()
{
    Q_D(QtAbstractStateGroup);
    for (int i = 0; i < d->exitActions.size(); ++i)
        QtStateActionPrivate::get(d->exitActions.at(i))->callExecute();
}

/*!
  Sets the restore policy of this state to \a restorePolicy. 
  
  The default restore policy is QtAbstractStateGroup::GlobalRestorePolicy.
*/
void QtAbstractStateGroup::setRestorePolicy(RestorePolicy restorePolicy)
{
    Q_D(QtAbstractStateGroup);
    d->restorePolicy = restorePolicy;
}

/*!
  Returns the restore policy for this state.
*/
QtAbstractStateGroup::RestorePolicy QtAbstractStateGroup::restorePolicy() const
{
    Q_D(const QtAbstractStateGroup);
    return d->restorePolicy;
}

#ifndef QT_NO_ANIMATION

/*!
  \overload addAnimatedTransition()

  Adds an animated transition from the current state to \a targetState for \a animation. 

  This function creates a QtSignalTransition for the \a sender and \a signal, and calls 
  addAnimatedTransition() with this transition object.
*/
QtAnimationState *QtAbstractStateGroup::addAnimatedTransition(const QObject *sender, const char *signal,
                                                            QtAbstractState *targetState,
                                                            QtAbstractAnimation *animation)
{
    return addAnimatedTransition(new QtSignalTransition(sender, signal), targetState, animation);
}

/*!
  Adds an animated transition from the current state to \a targetState. 
  
  The animated transition has an intermediate QtAnimationState which plays \a animation before 
  entering \a targetState. This QtAnimationState will be entered when \a transition is taken by the 
  state machine. When the animation has finished playing, the \a targetState will be entered.

  The new QtAnimationState object will become a child of \a targetState's parent state.

  \code
    QPushButton button;
    QtPropertyAnimation animation(&button, "geometry");
    animation.setEndValue(QRect(100, 100, 400, 400));

    QtStateMachine machine;

    QtState *s1 = new QtState();
    QtState *s2 = new QtState();

    QtTransition *transition = new QtTransition(MyEventType);
    s1->addAnimatedTransition(transition, s2, &animation);
  \endcode

  The function returns the new QtAnimationState. This state can be used if you want to add additional 
  transitions into or out from the animation state, and if you want to add additional animations.

  \sa QtAnimationState
*/
QtAnimationState *QtAbstractStateGroup::addAnimatedTransition(QtAbstractTransition *transition,
                                                            QtAbstractState *targetState, 
                                                            QtAbstractAnimation *animation)
{
    QtAnimationState *animState = new QtAnimationState(animation, targetState->parentState());
    animState->addFinishedTransition(targetState);
    addTransition(transition, animState);
    return animState;
}

#endif

QT_END_NAMESPACE
