/****************************************************************************
**
** This file is part of a Qt Solutions component.
** 
** Copyright (c) 2009 Nokia Corporation and/or its subsidiary(-ies).
** 
** Contact:  Qt Software Information (qt-info@nokia.com)
** 
** Commercial Usage  
** Licensees holding valid Qt Commercial licenses may use this file in
** accordance with the Qt Solutions Commercial License Agreement provided
** with the Software or, alternatively, in accordance with the terms
** contained in a written agreement between you and Nokia.
** 
** GNU Lesser General Public License Usage
** Alternatively, this file may be used under the terms of the GNU Lesser
** General Public License version 2.1 as published by the Free Software
** Foundation and appearing in the file LICENSE.LGPL included in the
** packaging of this file.  Please review the following information to
** ensure the GNU Lesser General Public License version 2.1 requirements
** will be met: http://www.gnu.org/licenses/old-licenses/lgpl-2.1.html.
** 
** In addition, as a special exception, Nokia gives you certain
** additional rights. These rights are described in the Nokia Qt LGPL
** Exception version 1.0, included in the file LGPL_EXCEPTION.txt in this
** package.
** 
** GNU General Public License Usage 
** Alternatively, this file may be used under the terms of the GNU
** General Public License version 3.0 as published by the Free Software
** Foundation and appearing in the file LICENSE.GPL included in the
** packaging of this file.  Please review the following information to
** ensure the GNU General Public License version 3.0 requirements will be
** met: http://www.gnu.org/copyleft/gpl.html.
** 
** Please note Third Party Software included with Qt Solutions may impose
** additional restrictions and it is the user's responsibility to ensure
** that they have met the licensing requirements of the GPL, LGPL, or Qt
** Solutions Commercial license and the relevant license of the Third
** Party Software they are using.
** 
** If you are unsure which license is appropriate for your use, please
** contact the sales department at qt-sales@nokia.com.
** 
****************************************************************************/



/*!
    \class QtAnimationGroup
    \brief The QtAnimationGroup is an abstract base class for group of animations.
    
    
    \preliminary

    QtAnimationGroup represents a group of animations, such as parallel or sequential,
    and lets you combine different animations into one. The group manages any animation
    that inherits QtAbstractAnimation. By combining groups, you can easily construct
    complex animation graphs.

    The QtAnimationGroup base class provides methods for adding and retrieving animations.
    Besides that, you can remove animations by calling remove(), and clear the animation
    group by calling clear(). You may keep track of changes in the group's animations by
    listening to QEvent::ChildAdded and QEvent::ChildRemoved events.

    QtAnimationGroup takes ownership of the animations it manages, and ensures that they are
    deleted when the animation group is deleted.

    \sa QtAbstractAnimation, QtAnimation, {The Animation Framework}
*/

#include "qanimationgroup.h"
#include <QtCore/qdebug.h>
#include <QtCore/qcoreevent.h>
#include "qanimationgroup_p.h"

QT_BEGIN_NAMESPACE


/*!
    Constructs a QtAnimationGroup.
    \a parent is passed to QObject's constructor.
*/
QtAnimationGroup::QtAnimationGroup(QObject *parent)
    : QtAbstractAnimation(*new QtAnimationGroupPrivate, parent)
{
}

/*!
    \internal
*/
QtAnimationGroup::QtAnimationGroup(QtAnimationGroupPrivate &dd, QObject *parent)
    : QtAbstractAnimation(dd, parent)
{
}

/*!
    Destroys the animation group. It will also destroy all its animations.
*/
QtAnimationGroup::~QtAnimationGroup()
{
    Q_D(QtAnimationGroup);
    foreach(QtAbstractAnimation *anim, d->animations) {
        //they well be deleted anyway.
        //this avoid calling removeAt from the destructor of QObject.
        anim->d_func()->group = 0;
    }
}

/*!
    Returns a pointer to the animation at \a index in this group. This
    function is useful when you need access to a particular animation.  \a
    index is between 0 and count() - 1.

    \sa count(), animations()
*/
QtAbstractAnimation *QtAnimationGroup::at(int index) const
{
    Q_D(const QtAnimationGroup);
    if (index < 0 || index >= d->animations.size())
        return 0;
    return d->animations.at(index);
}

/*!
    Returns the animations that are managed by this group.

    \sa add(), count(), at()
*/
QList<QtAbstractAnimation *> QtAnimationGroup::animations() const
{
    Q_D(const QtAnimationGroup);
    return d->animations;
}

/*!
    Returns the number of animations managed by this group.

    \sa animations(), add(), at()
*/
int QtAnimationGroup::count() const
{
    Q_D(const QtAnimationGroup);
    return d->animations.size();
}

/*!
    Adds \a animation to this group. This will call animationInsertedAt.

    \sa animationInsertedAt()
*/
void QtAnimationGroup::add(QtAbstractAnimation *animation)
{
    Q_D(QtAnimationGroup);

    d->animations.append(animation);
    if (QtAnimationGroup *oldGroup = animation->d_func()->group)
        oldGroup->remove(animation);
    animation->d_func()->group = this;
    animation->setParent(this);
    animationInsertedAt(d->animations.count() - 1);
}

/*!
    Removes \a animation from this group. The ownership of \a animation is
    transferred to the caller.

    \sa add()
*/
void QtAnimationGroup::remove(QtAbstractAnimation *animation)
{
    Q_D(QtAnimationGroup);
    if (!animation) {
        qWarning("QtAnimationGroup::remove: cannot remove null animation");
        return;
    }
    int index = d->animations.indexOf(animation);
    if (index == -1) {
        qWarning("QtAnimationGroup::remove: animation is not part of this group");
        return;
    }

    removeAt(index);
}

/*!
    Removes the animation at \a index from this animation group. The ownership
    of the animation is transferred to the caller. 
    This will call animationRemovedAt.

    \sa add(), animationRemovedAt()
*/
void QtAnimationGroup::removeAt(int index)
{
    Q_D(QtAnimationGroup);
    if (index < 0 || index >= d->animations.size()) {
        qWarning("QtAnimationGroup::remove: no animation at index %d", index);
        return;
    }
    QtAbstractAnimation *animation = d->animations.at(index);
    animation->d_func()->group = 0;
    d->animations.removeAt(index);
    animationRemovedAt(index);
}

/*!
    This virtual method is called whenever an animation is added to
    the group at index \a index.

    \sa add()
*/

void QtAnimationGroup::animationInsertedAt(int index)
{
    Q_UNUSED(index);
}

/*!
    This virtual method is called whenever an animation is removed from
    the group at index \a index. The animation is no more listed when this
    method is called.

    \sa remove(), removeAt()
*/

void QtAnimationGroup::animationRemovedAt(int index)
{
    Q_UNUSED(index);
}

/*!
    Removes and deletes all animations in this animation group, and resets the current
    time to 0.

    \sa add(), remove()
*/
void QtAnimationGroup::clear()
{
    Q_D(QtAnimationGroup);
    qDeleteAll(d->animations);
    d->animations.clear();
    setCurrentTime(0);
}

/*!
    \reimp
*/
bool QtAnimationGroup::event(QEvent *event)
{
    if (event->type() == QEvent::ChildAdded) {
        QChildEvent *childEvent = static_cast<QChildEvent *>(event);
        if (QtAbstractAnimation *a = qobject_cast<QtAbstractAnimation *>(childEvent->child())) {
            if (a->group() != this)
                add(a);
        }
    } else if (event->type() == QEvent::ChildRemoved) {
        QChildEvent *childEvent = static_cast<QChildEvent *>(event);
        if (QtAbstractAnimation *a = qobject_cast<QtAbstractAnimation *>(childEvent->child())) {
            // ### you can only rely on the child being a QObject.
            // ### This is because in the QEvent::ChildRemoved case it might have already been destructed.
            if (a->group() != this)
                remove(a);
        }
    }
    return QtAbstractAnimation::event(event);
}

QT_END_NAMESPACE

#include "moc_qanimationgroup.cpp"

